program aqmonitor;

{$mode objfpc}{$H+}

uses
    {$IFDEF UNIX}{$IFDEF UseCThreads}
    cthreads,
    {$ENDIF}{$ENDIF}
    {$IFDEF UNIX}
    BaseUnix,
    {$ENDIF}
    Classes,
    DateUtils,
    db,
    sqldb,
    sqlite3conn,
    synaser,
    SysUtils;

// -------------------------------------------------------------------------
// Global Constants

const
    dbName                              =   'aqmonitor.db';
    duration                            =   5;                              // time between db writes in minutes
    erMsgChecksum                       =   'Checksum for sensor data failed';
    erMsgFrameLen                       =   'Frame length was incorrect';
    erMsgTimeout                        =   'Timeout waiting for sensor data';
    maxChecksumCount                    =   10;
    maxInsertCount                      =   10;                             // max # of successive inserts to fail
    maxPacketHistory                    =   360;                            // max # of packets to keep in history
    maxRetries                          =   9;                              // max # of times to RETRY insert due to locked db
    maxTimeoutCount                     =   10;
    timeOut                             =   30;                             // max seconds to wait for sensor to respond

// -------------------------------------------------------------------------
// Global types / records

type

aqPacketT                               =   record
    hdr1                                :   byte;
    hdr2                                :   byte;
    frameLen                            :   word;
    pm10Std                             :   word;
    pm25Std                             :   word;
    pm100Std                            :   word;
    pm10Env                             :   word;
    pm25Env                             :   word;
    pm100Env                            :   word;
    particles03um                       :   word;
    particles05um                       :   word;
    particles10um                       :   word;
    particles25um                       :   word;
    particles50um                       :   word;
    particles100um                      :   word;
    filler                              :   word;
    checksum                            :   word;
    end; // aqPacketT

EChecksum                               =   class(Exception);
EFrameLen                               =   class(Exception);
EInsertFailed                           =   class(Exception);
ETimeout                                =   class(Exception);

histPacketT                             =   record
    capTime                             :   TDateTime;                      // capture time
    packet                              :   aqPacketT;
    end;

HistoryT                                =   record                          // circular list of packet history
    max                                 :   integer;                        // total elements
    len                                 :   integer;                        // current # elements
    next                                :   integer;                        // location for next add (oldest packet)
    Packets                             :   array of histPacketT;
    end; // aqPacketT

// -------------------------------------------------------------------------
// global variables

var
    abortRequested                      :   boolean         = false;
    comCB                               :   TBlockSerial;
    connector                           :   TSQLConnector;
    dbCB                                :   TSQLConnection;
    errChecksumCount                    :   integer;
    errInsertCount                      :   integer;
    errTimeoutCount                     :   integer;
    packet                              :   aqPacketT;
    history                             :   HistoryT;
    requestedPort                       :   string;
    avgTimer                            :   TDateTime;

// -------------------------------------------------------------------------
// Forward Declarations

procedure dbAdd(
    duration                            :   integer;
    packet                              :   aqPacketT
    );
    forward;

procedure dbClose;
    forward;

procedure dbOpen;
    forward;

function getByte(
    comCB                               :   TBlockSerial
    )                                   :   byte;
    forward;

function getComPort                     :   string
    forward;

function getPacket(
    comCB                               :   TBlockSerial
    )                                   :   aqPacketT;
    forward;

function getWord(
    comCB                               :   TBlockSerial;
    var checksum                        :   word
    )                                   :   word;
    forward;

procedure handleSigInt(
    aSignal                             : LongInt
    );
    cdecl;
    forward;

procedure historyAdd(
    var history                         :   historyT;
    packet                              :   aqPacketT
    );
    forward;

function historyAvg(
    seconds                             :   integer;                        // # of seconds of hist to average from 'now'
    history                             :   HistoryT
    )                                   :   aqPacketT;
    forward;

procedure historyDump(
    history                             :   historyT
    );
    forward;

procedure historyInit(
    var history                         :   historyT
    );
    forward;

function initComPort(
    comPort                             :   string
    )                                   :   TBlockSerial;
    forward;

procedure printPacket(
    packet                              :   aqPacketT
    );
    forward;

// -------------------------------------------------------------------------
// add observation to db

procedure dbAdd(
    duration                            :   integer;
    packet                              :   aqPacketT
    );

var
    q                                   :   TSQLQuery;
    retries                             :   integer;
    s                                   :   string;
    trans                               :   TSQLTransaction;

begin

with packet do begin
    s := 'insert into observations (' +
            'date, duration, pm10Std, pm25Std, pm100Std, pm10Env, pm25Env, pm100E, ' +
            'part03, part05, part10, part25, part50, part100' +
            ')' +
        'values(' +
            '"' + FormatDateTime('yy-mm-dd hh:nn:ss', now) + '", ' +
            inttostr(duration) + ', ' +
            inttostr(pm10Std) + ', ' +
            inttostr(pm25Std) + ', ' +
            inttostr(pm100Std) + ', ' +
            inttostr(pm10Env) + ', ' +
            inttostr(pm25Env) + ', ' +
            inttostr(pm100Env) + ', ' +
            inttostr(particles03um) + ', ' +
            inttostr(particles05um) + ', ' +
            inttostr(particles10um) + ', ' +
            inttostr(particles25um) + ', ' +
            inttostr(particles50um) + ', ' +
            inttostr(particles100um) +
        ')';
    end; // with
//writeln(s);

try try
    trans             := TSQLTransaction.Create(nil);
    trans.DataBase    := dbCB;
    trans.StartTransaction;

    q                 := TSQLQuery.Create(nil);
    q.DataBase        := dbCB;
    q.SQL.Text        := s;

    // attempt to insert record. If dblocked occurs (error 5), then retry until
    // succeeds or # of retries are exceeded
    retries := 0;
    while true do begin
        try // insert
            q.ExecSQL;
        except
            on e: ESQLDatabaseError do begin                                // db error?
                if e.ErrorCode = 5 then begin                               // db locked error?
                    retries := retries + 1;
                    if retries > maxRetries then begin                      // # of retries exceeded?
                        writeln('Insert failed due to dblock after ', retries,
                            ' attempts.');
                        raise EInsertFailed.Create                          // pass dblocked error to caller
                            ('Insert retries exceeded due to dblock');
                        end;
                    sleep(retries * 1000);                                  // sleep 1000, 2000, 3000 ms ...
                    continue;                                               // restart while loop
                    end
                else
                    raise                                                   // unexpected error, return it to sender
                end; // on
            end; //try insert
        break;
        end; // while

except
    on e: EDatabaseError do begin
        writeln('Error: ' + e.Message);
        writeln('insert failed.');
        writeln(s);
        raise;
        end;
    end; // try except;

finally
    q.close;
    q.Free;
    trans.Commit;
    trans.Free;
    end; // try finally

end;

// -------------------------------------------------------------------------
// close database

procedure dbClose;

begin

try try
    dbCB.Connected := false;

except
    on e: EDatabaseError do begin
        writeln('Error: ', e.Message);
        writeln('Unable to close db ', dbName);
        raise;
        end;
    end // except

finally
    dbCB.Free;
    connector.Free;
    end; // finally

end; // dbClose

// -------------------------------------------------------------------------
// open database

procedure dbOpen;

var
    trans                               : TSQLTransaction;

begin

if not FileExists(dbname) then begin
    raise EDatabaseError.create('Missing db file: ' + dbname);
    end;

try
    connector               := TSQLConnector.Create(nil);
    connector.connectorType := 'SQLite3';
    dbCB                    := TSQLite3Connection.Create(nil);

    dbCB.Params.Add       ('foreign_keys=on');
    dbCB.HostName         := 'localhost';
    dbCB.DatabaseName     := dbName;
    dbCB.UserName         := '';
    dbCB.Password         := '';
    dbCB.Connected        := true;

except
    on e: exception do begin
        writeln('Error: ', e.Message);
        writeln('Unable to open db ', dbName);
        raise;
        end;
    end;

end; //dbOpen;

// -------------------------------------------------------------------------
// get next waiting byte.

function getByte(
    comCB                               :   TBlockSerial
    )                                   :   byte;

var
    timer                               :   TDateTime;

begin

timer := Now;
while comCB.WaitingData = 0 do begin                                        // wait for data
    if SecondsBetween(now, timer) >= timeOut then begin
        // didn't get a response from sensor, so gen a timeout exception.
        raise ETimeout.Create(erMsgTimeout);
        end;
    Sleep(10);
    end; // while

result := comCB.RecvByte(50);                                               // get next waiting byte

end; // getByte

// -------------------------------------------------------------------------
// get a valid com port, either from the 1st parameter of the command line
// or by asking users.

function getComPort                     :   string;

var
    s                                   :   string;

begin

if ParamCount > 0 then begin
    // parameter was supplied so check it
    if ParamCount > 1 then begin
        writeln;
        writeln('Monitors Air Quality sensor. Command Format is:');
        writeln;
        writeln('aqmonitor [comport]');
        writeln;
        writeln('where ');
        writeln;
        writeln('  comport     comport name such as COM1 or /dev/ttyUSB0. If omitted,');
        writeln('              then program will prompt for it.');
        writeln;
        halt(1);
        end;
    result := ParamStr(1);
    end
else begin
    write('Enter comm port name use: ');
    readln(s);
    if (length(s) = 0) then begin
        writeln('No com port selected, exiting.');
        halt;
        end;
    result := s;
    end;

end; // getComPort

// -------------------------------------------------------------------------
// Read bytes from com port until a full packet has been created.

function getPacket(
    comCB                               :   TBlockSerial
    )                                   :   aqPacketT;

var
    calcChecksum                        :   word;
    finalChecksum                       :   word;

begin

while true do begin
    // read until we get a packet header
    while getbyte(comCB) <> $42 do;

    // next byte must be $4D
    if getbyte(comCB) = $4D then
        break;

    // $42 $4D not found, go back to searching

    end; // while;

with getPacket do begin
    hdr1              := $42;
    hdr2              := $4D;

    calcChecksum := hdr1 + hdr2;

    frameLen          := getword(comCB, calcChecksum);
    if frameLen <> 28 then
        raise EFrameLen.Create(erMsgFrameLen);

    pm10Std           := getword(comCB, calcChecksum);
    pm25Std           := getword(comCB, calcChecksum);
    pm100Std          := getword(comCB, calcChecksum);
    pm10Env           := getword(comCB, calcChecksum);
    pm25Env           := getword(comCB, calcChecksum);
    pm100Env          := getword(comCB, calcChecksum);
    particles03um     := getword(comCB, calcChecksum);
    particles05um     := getword(comCB, calcChecksum);
    particles10um     := getword(comCB, calcChecksum);
    particles25um     := getword(comCB, calcChecksum);
    particles50um     := getword(comCB, calcChecksum);
    particles100um    := getword(comCB, calcChecksum);
    filler            := getword(comCB, calcChecksum);

    finalChecksum     := calcChecksum;
    checksum          := getword(comCB, calcChecksum);

    if finalChecksum <> checksum then begin
        // checksums don't match, throw exception
        raise EChecksum.Create(erMsgChecksum);
        end;

    end; // with

end; // getPacket

// -------------------------------------------------------------------------
// get next 2 bytes as a single (16bit) word. The most significant byte is read
// first.
// The checksum of the 2 bytes is added to <checksum>

function getWord(
    comCB                               :   TBlockSerial;
    var checksum                        :   word
    )                                   :   word;

var
    lsb                                 :   byte;
    msb                                 :   byte;

begin

msb    := getByte(comCB);
lsb    := getByte(comCB);
result := (msb << 8) OR lsb;

checksum := checksum + msb + lsb;

end; // getWord

// -------------------------------------------------------------------------
// If user presses control-c, this is called

procedure handleSigInt(
    aSignal                             : LongInt
    );
    cdecl;

begin

writeln;
writeln('User requested program termination.');
writeln;
abortRequested := true;

end; // handleSigInt
 
// -------------------------------------------------------------------------
// Add packet to history array

procedure historyAdd(
    var history                         : historyT;
    packet                              : aqPacketT
    );

begin

with history do begin
    packets[next].capTime := now;
    packets[next].packet  := packet;
    next := (next + 1) mod max;
    if len < max then
        len := len + 1;
    end; // with

end; // historyAdd

// -------------------------------------------------------------------------
// compute moving average of values in history for the past n seconds.
// If there are less than n seconds of history, you will get the moving
// average for the data available.

function historyAvg(
    seconds                             :   integer;                        // # of seconds to average
    history                             :   HistoryT
    )                                   :   aqPacketT;

var
    dataPoints                          :   integer;
    dw                                  :   record
        pm10Std			                :	dword;
        pm25Std			                :	dword;
        pm100Std			            :	dword;
        pm10Env			                :	dword;
        pm25Env			                :	dword;
        pm100Env			            :	dword;
        particles03um			        :	dword;
        particles05um			        :	dword;
        particles10um			        :	dword;
        particles25um			        :	dword;
        particles50um			        :	dword;
        particles100um			        :	dword;
        end;
    n                                   :   integer;
    time                                :   TDateTime;

begin

// determine where to start (find oldest record as <n>)
with history do begin
    if len < max then
        n := 0
    else
        n := next;
    end; // with

with dw do begin
    pm10Std        := 0;
    pm25Std        := 0;
    pm100Std       := 0;
    pm10Env        := 0;
    pm25Env        := 0;
    pm100Env       := 0;
    particles03um  := 0;
    particles05um  := 0;
    particles10um  := 0;
    particles25um  := 0;
    particles50um  := 0;
    particles100um := 0;
    end; // with

time                      := now;
dataPoints                := 0;

// move thru array summing all values until time exceeded or end hit

repeat

    if SecondsBetween(time, history.Packets[n].capTime) <= seconds then begin
        // packet within time range, so add it to avg
        with history.Packets[n].packet do begin
            dw.pm10Std        := dw.pm10Std        + pm10Std;
            dw.pm25Std        := dw.pm25Std        + pm25Std;
            dw.pm100Std       := dw.pm100Std       + pm100Std;
            dw.pm10Env        := dw.pm10Env        + pm10Env;
            dw.pm25Env        := dw.pm25Env        + pm25Env;
            dw.pm100Env       := dw.pm100Env       + pm100Env;
            dw.particles03um  := dw.particles03um  + particles03um;
            dw.particles05um  := dw.particles05um  + particles05um;
            dw.particles10um  := dw.particles10um  + particles10um;
            dw.particles25um  := dw.particles25um  + particles25um;
            dw.particles50um  := dw.particles50um  + particles50um;
            dw.particles100um := dw.particles100um + particles100um;
            end; // with
        dataPoints := dataPoints + 1;
        end;

    n := (n + 1) mod history.max;

    until n = history.next;

// divide to get avg
with historyAvg do begin
    pm10Std        := round(dw.pm10Std        / dataPoints);
    pm25Std		   := round(dw.pm25Std        / dataPoints);
    pm100Std	   := round(dw.pm100Std       / dataPoints);
    pm10Env		   := round(dw.pm10Env        / dataPoints);
    pm25Env		   := round(dw.pm25Env        / dataPoints);
    pm100Env	   := round(dw.pm100Env       / dataPoints);
    particles03um  := round(dw.particles03um  / dataPoints);
    particles05um  := round(dw.particles05um  / dataPoints);
    particles10um  := round(dw.particles10um  / dataPoints);
    particles25um  := round(dw.particles25um  / dataPoints);
    particles50um  := round(dw.particles50um  / dataPoints);
    particles100um := round(dw.particles100um / dataPoints);
    end; // with

end; // historyAvg

// -------------------------------------------------------------------------
// formatted dump of history array

procedure historyDump(
    history                             :   historyT
    );

var
    n                                   :   integer;

begin

with history do begin

    writeln;
    writeln('max:  ', max);
    writeln('len:  ', len);
    writeln('next: ', next);

    if len = 0 then begin
        writeln('History array is empty.');
        exit;
        end;

    if len < max then
        n := 0
    else
        n := next;
    end; // with

writeln(
    '--date-- --time--', ' ',
    '1.0S':5, ' ', '2.5S':5, ' ', '10.0S':5, ' ',
    '1.0E':5, ' ', '2.5E':5, ' ', '10.0E':5, ' ',
    '0.3P':5, ' ', '0.5P':5, ' ', '1.0P':5,  ' ',
    '2.5P':5, ' ', '5.0P':5, ' ', '10.0P':5
    );
writeln('--------------------------------------------------------',
    '----------------------------------');

repeat

    with history.Packets[n].packet do begin
        writeln(
            FormatDateTime('yy-mm-dd hh:nn:ss', history.Packets[n].capTime), ' ',
            pm10Std:5, ' ', pm25Std:5, ' ', pm100Std:5, ' ',
            pm10Env:5, ' ', pm25Env:5, ' ', pm100Env:5, ' ',
            particles03um:5, ' ', particles05um:5, ' ', particles10um:5, ' ',
            particles25um:5, ' ', particles50um:5, ' ', particles100um:5
            );
        end; // with

    n := (n + 1) mod history.max;

    until n = history.next;

end; // historyDump

// -------------------------------------------------------------------------
// Initialize the history array

procedure historyInit(
    var history                         :   historyT
    );

begin

with history do begin
    max  := maxPacketHistory;
    len  := 0;
    next := 0;
    SetLength(packets, max);
    end;

end; // historyInit

// -------------------------------------------------------------------------
// Initialize communications serial port

function initComPort(
    comPort                             :   string
    )                                   :   TBlockSerial;

var
	baud								:   string;

begin


baud := '9600';

initComPort :=TBlockSerial.Create;

initComPort.RaiseExcept := true;

try
    initComPort.Connect(comPort);
    initComPort.config(strToInt(baud), 8, 'N', SB1, false, false);

except
    on e: exception do begin
        writeln(e.Message);
        writeln('Unable to open communications port. Terminating.');
    	raise;
        end;
    end; // try

end; // initComPort

// -------------------------------------------------------------------------
// Format a packet for viewing.

procedure printPacket(
    packet                              :   aqPacketT
    );

const
    lines                               :   integer         = 0;

begin

if lines mod 10 = 0 then begin
    writeln(
        '--date-- --time--', ' ',
        '1.0S':5, ' ', '2.5S':5, ' ', '10.0S':5, ' ',
        '1.0E':5, ' ', '2.5E':5, ' ', '10.0E':5, ' ',
        '0.3P':5, ' ', '0.5P':5, ' ', '1.0P':5,  ' ',
        '2.5P':5, ' ', '5.0P':5, ' ', '10.0P':5
        );
    writeln('--------------------------------------------------------',
        '----------------------------------');
    end;

with packet do begin
    writeln(
        FormatDateTime('yy-mm-dd hh:nn:ss', now), ' ',
        pm10Std:5, ' ', pm25Std:5, ' ', pm100Std:5, ' ',
        pm10Env:5, ' ', pm25Env:5, ' ', pm100Env:5, ' ',
        particles03um:5, ' ', particles05um:5, ' ', particles10um:5, ' ',
        particles25um:5, ' ', particles50um:5, ' ', particles100um:5
        );
    end; // with

lines := lines + 1;

end; // printPacket

// -------------------------------------------------------------------------

begin // outerblock

writeln;
writeln('aqMonitor: Air Quality Monitor. Compiled: ',
    {$I %DATE%} + ' ' + {$I %TIME%}, '.');
writeln;


try

    requestedPort := getComPort;
    comCB         := initComPort(requestedPort);

    // trap for signals that could cause abnormal termination so the com port
    // can be properly closed
    if FpSignal(SIGINT, @HandleSigInt) = signalhandler(SIG_ERR) then begin
        Writeln('Failed to install signal error: ', fpGetErrno);
        Halt(1);
        end;
    if FpSignal(SIGHUP, @HandleSigInt) = signalhandler(SIG_ERR) then begin
        Writeln('Failed to install signal error: ', fpGetErrno);
        Halt(1);
        end;
    if FpSignal(SIGTERM, @HandleSigInt) = signalhandler(SIG_ERR) then begin
        Writeln('Failed to install signal error: ', fpGetErrno);
        Halt(1);
        end;

    dbOpen;

    //while not abortRequested do begin
    //    Initialize(packet);
    //    dbAdd(0,packet);
    //    sleep(5000);
    //    end;
    //exit;

    initialize(history);
    historyInit(history);

    writeln;
    writeln('Begin monitoring of sensor.');
    writeln;

    avgTimer         := now;
    errChecksumCount := 0;
    errInsertCount   := 0;
    errTimeoutCount  := 0;

    while not abortRequested do begin
        try
            packet           := getPacket(comCB);                           // get packet from sensor
            errChecksumCount := 0;
            errTimeoutCount  := 0;
        except
            on e: EChecksum do begin                                        // handle checksum error
                errChecksumCount := errChecksumCount + 1;
                if errChecksumCount > maxChecksumCount then begin
                    writeln('Checksum error count exceeded ', maxChecksumCount);
                    raise;
                    end;
                end;
            on e: ETimeout do begin                                         // handle timeout error
                errTimeoutCount := errTimeoutCount + 1;
                if errTimeoutCount > maxTimeoutCount then begin
                    writeln('Timeout error count exceeded ', maxTimeoutCount);
                    raise;
                    end;
                end;
            end; // try except

        //printPacket(packet);
        historyAdd(history, packet);                                        // add packet to history
        if SecondsBetween(now, avgTimer) >= duration * 60 then begin        // time to avg & write to db?
            //historyDump(history);
            packet := historyAvg(duration * 60, history);                   // average last d*60 seconds of packets
            try
                dbAdd(duration, packet);                                    // write this 'average' pkt to db
                errInsertCount := 0;                                        // good dbAdd, so clear error count
            except
                on e: EInsertFailed do begin                                // EInsertFailed ?
                    errInsertCount := errInsertCount + 1;                   // inc insert error counter
                    if errInsertCount > maxInsertCount then begin           // too many successive inserts failed?
                        writeln('Insert error count exceeded: ', maxInsertCount);
                        raise;
                        end;
                    end;
                end; // try
            printPacket(packet);                                            // output to stdlist
            avgTimer := now;                                                // set time for next timeout
            end;
        end; // while
finally
    writeln('Closing database.');
    dbClose;

    writeln('Closing COM port.');
    comCB.CloseSocket;
    comCB.Destroy;
    writeln;
    end;
end.

